/**
 * The Paper Tape Project -- Font subproject
 * PaperTapeFont C Library / Stand alone frontend
 *
 * This file contains library routines (see papertapefont.h) for
 * reading and writing "papertape font" files, as well as a
 * powerful cross platform working stand alone program.
 *
 * To compile the CLI program, define STANDALONE, e.g.
 * with
 * 
 *   gcc papertapefont.c -DSTANDALONE=1
 *
 * See the document "FONT_FILES.txt" in the source code directory
 * for an introducion and manual.
 *
 * Initially written in a few hours (500 lines code) between
 * 02:00 and 06:00 at 04.09.2008
 * Improved and multiple times rewritten the next day to
 * 1000 lines ;-)
 *
 * (c) 2008 Sven Kppel
 *
 * This program is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, see
 * http://www.gnu.org/licenses/.
 *
 **/

#include "papertapefont.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

byte_t papertape_font_line2byte(const char *line);
void papertape_font_byte2line(const byte_t byte, FILE *write);

/**
 * Allocate a new PAPERTAPE_FONT object, without any character.
 * You can use papertape_font_new_from_file() if you want to create
 * a PAPERTAPE_FONT object from file or use papertape_font_import_from_file()
 * or papertape_font_import_from_schematics() to fill your PAPERTAPE_FONT
 * object.
 **/
// implemented as a Macro due to simplicity.

/**
 * Load a new papertape font from the file called "filename". This
 * is some more luxurious than calling the single functions.
 * @return allocated font object or NULL if not possible. 
 **/
PAPERTAPE_FONT* papertape_font_new_from_file(const char* filename) {
	FILE* file;
	PAPERTAPE_FONT* font = papertape_font_new();
	if(!font)
		return NULL;

	file = fopen(filename, "r");
	if(!file) {
		perror("Opening papertape font file");
		return NULL;
	}

	papertape_font_load_file(font, file);
	return font;
}

/**
 * This will load all characters found in the filehandle (which must be readable).
 * Characters loaded here will overwrite previous characters, if present.
 **/
void papertape_font_load_file(PAPERTAPE_FONT* font, FILE* fh) {
	#define MAX_LINE   100
	#define MAX_BYTES  30
	#define MAX_NAME   30

	// Buffer for the line string currently read in
	char line[MAX_LINE];
	// Buffer for the name string currently read in
	char name[MAX_NAME] = "";
	// Buffer for last name string read in
	char last_name[MAX_NAME] = "";
	// Buffer for the bytes currently read in
	byte_t bytes[MAX_BYTES];
	// Real number of bytes already read in
	int bytes_n = 0;
	// the current ascii_id
	char ascii_id = 0;
	// the last ascii_id
	char last_ascii_id = 0;
	// if we haven't already read in one byte from the new
	// character, this is set to 1. This will force reading in the
	// current byte.
	int new_character = 1;

	// for each character package...
	while(fgets(line, MAX_LINE, fh) != NULL) {
		/*
		 * the layout of a typical   *  the layout of a typical long
		 * ASCII line:               *  special line:
		 *                           *
		 *     |123.45678|x          *  |123.45678| =named_thing  comments
		 *     012345678901          *  01234567890123456789012345678...
		 *                           *             ||
		 * length: 12 characters.    *    12 chars-^^- char no. 12
		 *                           *
		 */

		/* remove "\n" lines at end. We don't need them. */
		if(line[strlen(line)-1] == '\n')
			line[strlen(line)-1] = '\0';
	
		/* skip lines which are not a paper tape or  which have no id */
		if(line[0] != PAPERTAPE_FONT_BORDER || strlen(line) < 12)
			continue;

		ascii_id = line[11];
		//fprintf(stderr, "Read in line ascii_id 0x%x (%c)\n", ascii_id, ascii_id);

		// strlen: to avoid segfaults
		if(strlen(line) > (11 + strlen(PAPERTAPE_FONT_SPECIAL_ID)) &&
		   strncmp(line+11, PAPERTAPE_FONT_SPECIAL_ID, strlen(PAPERTAPE_FONT_SPECIAL_ID)) == 0) {
			// we have a special value with a full name
			ascii_id = 0;
			int z;
			// copy name
			strncpy(name, &line[11+strlen(PAPERTAPE_FONT_SPECIAL_ID)], MAX_NAME);
			// truncate name to only one word
			for(z=0; !(name[z] == ' ' || name[z] == '\0' || name[z] == '\t'); z++);
			name[z] = '\0';
		}

		// if we are still at the same byte like the last one...
		if(ascii_id == last_ascii_id || //new_character ||
		   ( ascii_id == 0 && (strcmp(name, last_name) == 0) )
		  ) {
			//fprintf(stderr, "Save line for 0x%x as no. %i\n", ascii_id, bytes_n+1);
			// ...then add this byte to the current char byte array
			if(bytes_n < MAX_BYTES)
			     // only store if enough place in bytes[] array
			     bytes[bytes_n++] = papertape_font_line2byte(line);
			// first byte stored, no more new character
			new_character = 0;
		} else {
			// we are at a new char, so let's push the last char to the list
			if(last_ascii_id == 0)
				papertape_font_set_special(font, last_name, bytes, bytes_n);
			else
				papertape_font_set_char(font, last_ascii_id, bytes, bytes_n);

			// store first byte from new character
			bytes[0] = papertape_font_line2byte(line);
			bytes_n = 1;
			// and tell the next loop that it has to read in.
			new_character = 1;
		}

		// Don't force to read in any more.
		//if(last_ascii_id == ascii_id)
		//	new_character = 0;
		last_ascii_id = ascii_id;
		strcpy(last_name, name);
	} // while file
	
	// we are done, and there's still one character left to push to the list
	if(last_ascii_id == 0)
		papertape_font_set_special(font, last_name, bytes, bytes_n);
	else
		papertape_font_set_char(font, last_ascii_id, bytes, bytes_n);

	// rewind the file handle (if writing is called on the same file, etc.)
	rewind(fh);
}

/**
 * Load Papertape font characters from binary array in order which is given
 * by the "schematics" string.
 *
 **/
void papertape_font_load_schematics(PAPERTAPE_FONT* font, const char *schematics, const byte_t *contents, int content_len) {
	int schematics_len = strlen(schematics);
	int schematics_x;
	int content_offset, content_start, content_end;
	for(schematics_x=0, content_offset=0; schematics_x < schematics_len; schematics_x++) {
		// find start of current character
		for(content_start = content_offset; content_start < content_len &&
			contents[content_start] == 0; content_start++);
		// find end of current character
		for(content_end = content_start; content_end < content_len &&
			contents[content_end] != 0; content_end++);
		// make new character between content_start and content_end
		papertape_font_set_char(font, schematics[schematics_x],
			contents + content_start, content_end - content_start);
		// save offset for next character
		content_offset = content_end + 1;
	}
}



/**
 * Parse a line to a byte. This is only used internally.
 * Yeah, and it would be more performant to define this as a Macro, but that
 * can be performed using gcc -O2, too.
 **/
byte_t papertape_font_line2byte(const char *line) {
	byte_t value = 0;
	#define isPunched(pos, add) if(line[pos] != ' ') { value += add; }
	/* left border */
	isPunched(1 ,  1);
	isPunched(2,   2);
	isPunched(3,   4);
	/* feed hole */
	isPunched(5,   8);
	isPunched(6,  16);
	isPunched(7,  32);
	isPunched(8,  64);
	isPunched(9, 128);
	/* right border */
	return value;
}

/**
 * This is the invert function of papertape_font_line2byte. It writes to
 * file handle "write" a line, but without Newline!
 *
 **/
void papertape_font_byte2line(const byte_t byte, FILE *write) {
	#define paintPunched(check) fputc((byte & check) ?\
		PAPERTAPE_FONT_WRITE_LOGICAL_1 : ' ', write)
	fputc(PAPERTAPE_FONT_BORDER, write); /* left border */
	paintPunched(  1); /* bit 1 */
	paintPunched(  2); /* bit 2 */
	paintPunched(  4); /* bit 3 */
	fputc(PAPERTAPE_FONT_FEED_HOLE, write); /* feed hole */
	paintPunched(  8); /* bit 4 */
	paintPunched( 16); /* bit 5 */
	paintPunched( 32); /* bit 6 */
	paintPunched( 64); /* bit 7 */
	paintPunched(128); /* bit 8 */
	fputc(PAPERTAPE_FONT_BORDER, write); /* right border */
}


/**
 * This function adds or changes special characters in the papertape font.
 * These characters are typically white spaces or word spacings.
 *
 * @param font The Papertape Font object, e.g. made by papertape_font_new_...()
 * @param name The Name ID
 * @param bytes Your byte array which is supposed to represent the character. The function copies it.
 * @param bytes_n The length of your byte array (counting from 1, of course)
 * @return 0 if successfully added, 1 otherwise. (It will never return 1 ;-) )
 **/
PAPERTAPE_FONT_CHAR* papertape_font_set_special(PAPERTAPE_FONT* font, const char *name, const byte_t* bytes, int bytes_n) {
	// allocate the new character structure
	PAPERTAPE_FONT_CHAR *buf;
	buf = (PAPERTAPE_FONT_CHAR*) malloc(sizeof(PAPERTAPE_FONT_CHAR));

	// copy all values to structure
	buf->ascii_id = 0x0;
	buf->bytes_n = bytes_n;
	buf->bytes = (byte_t*) malloc(bytes_n);
	memcpy(buf->bytes, bytes, bytes_n);
	buf->name = (char*) malloc(strlen(name));
	strcpy(buf->name, name);
	
	// Pack it at the very start
	buf->next = (font->next != NULL) ? font->next : NULL;
	font->next = buf;

	return buf;
}

/**
 * This is the main function how to add or change a character in the papertape_font.
 * Internally, the papertape_font works with a ascii ordered single linked list. So
 * there's no other write access to that list than this function.
 *
 * If you want to set/add characters with special long names, use papertape_font_set_special_char.
 *
 * @param font The Papertape Font object, e.g. made by papertape_font_new_...()
 * @param ascii_id The ASCII value of the character you want to add.
 * @param bytes Your byte array which is supposed to represent the character. The function copies it.
 * @param bytes_n The length of your byte array (counting from 1, of course)
 * @return A link to the generated character. You don't need it, anyway.
 **/
PAPERTAPE_FONT_CHAR* papertape_font_set_char(PAPERTAPE_FONT* font, char ascii_id, const byte_t* bytes, int bytes_n) {
	PAPERTAPE_FONT_CHAR *current;
	PAPERTAPE_FONT_CHAR *next;
	
	// allocate the new character structure
	PAPERTAPE_FONT_CHAR *buf;
	buf = (PAPERTAPE_FONT_CHAR*) malloc(sizeof(PAPERTAPE_FONT_CHAR));

	// copy all values to structure
	buf->ascii_id = ascii_id;
	buf->bytes_n = bytes_n;
	buf->bytes = (byte_t*) malloc(bytes_n);
	memcpy(buf->bytes, bytes, bytes_n);
	buf->name = NULL;
	buf->next = NULL;

	// Start going throught the list
	current = font; // Start with the font object as first list member.
	while(current->next != NULL) {
		if(current->next->ascii_id == ascii_id) {
			// the next one has the same ascii_id as we have (perhaps
			// we are 'c' and the next is 'c', too). So replace the next
			// one.
			next = current->next;  // copy the pointer somewhere else
			current->next = buf;   // replace the next pointer with our char
			free(next);            // delete the old next pointer.
			return buf;
		} else if(current->next->ascii_id > ascii_id) {
			// the next one has a bigger ascii_id as we have (perhapse
			// we are 'c' while the next is 'e'). So pack our character
			// right here, between current and next.
			next = current->next;  // copy the pointer somewhere else
			current->next = buf;   // replace the next pointer with our char
			buf->next = next;      // pack the next pointer after our pointer
			return buf;
		} else {
			// the next one has a lower ascii_id as we have, so we have
			// to go on throught the list
			current = current->next;
		}
	}

	// we have reached the end of the list and haven't found any ascii_id biggger
	// or equal to ours. So we simply add our character to the list.
	current->next = buf;
	return buf;
}

/**
 * This is an auxillary function to access the linked list. It's not supposed to be used
 * from outside -- use highlevel functions like set_char or get_label.
 * @returns Link inside the list, NULL if not found.
 **/
PAPERTAPE_FONT_CHAR* papertape_font_low_get_char(PAPERTAPE_FONT* font, char ascii_id) {
	PAPERTAPE_FONT_CHAR* current;
	current = font;
	while(current->next != NULL) {
		if(current->next->ascii_id == ascii_id) {
			// we have found the right character entry.
			return current->next;
		} else
			current = current->next;
	}
	return NULL;
}

/**
 * This is just like papertape_font_low_get_char, but for special characters.
 * Ah, one special thing: Special characters are always valid, even if not defined.
 **/
PAPERTAPE_FONT_CHAR* papertape_font_low_get_special(PAPERTAPE_FONT* font, const char* name) {
	PAPERTAPE_FONT_CHAR* current;
	byte_t default_bytes[] = { 0 };
	current = font;
	while(current->next != NULL) {
		if(current->next->name != NULL &&
		   strcmp(current->next->name, name) == 0) {
			// we have found the right character entry.
			return current->next;
		} else
			current = current->next;
	}
	// no special character found. We'll generate one.
	return papertape_font_set_special(font, name, default_bytes, 1);
}


/**
 * I don't know why one wanted to delete a character, but by using this function you can
 * do exactly that.
 * @returns 1 if successful deleted, 0 if not found in list.
 **/
int papertape_font_del_char(PAPERTAPE_FONT* font, char ascii_id) {
	// we cannot use papertape_font_low_get_char, because this would need
	// a double linked list.
	PAPERTAPE_FONT_CHAR* current;
	current = font;
	while(current->next != NULL) {
		if(current->next->ascii_id == ascii_id) {
			// we have found the right character entry.
			if(current->next->next != NULL) {
				// we are somewhere in the list, so we have to rewrite
				// the "next" pointer.
				current->next = current->next->next;
			}
			free(current->next);
			return 1;
		} else
			current = current->next;
	}
	return 0;
}

/**
 * This is the same like for normal characters, only for special characters.
 **/
int papertape_font_del_special(PAPERTAPE_FONT* font, const char* name) {
	PAPERTAPE_FONT_CHAR* current;
	current = font;
	while(current->next != NULL) {
		if(current->next->name != NULL &&
		   strcmp(current->next->name, name) == 0) {
			// we have found the right character entry.
			if(current->next->next != NULL) {
				// we are somewhere in the list, so we have to rewrite
				// the "next" pointer.
				current->next = current->next->next;
			}
			free(current->next);
			return 1;
		} else
			current = current->next;
	}
	return 0;
}

/**
 * This is the main function from the papertape_font package. Using this function, you
 * can render any ASCII string to a papertape label, using the current font.
 *
 * The return byte array will be malloced, and you'll have after calling the only
 * link to that array. So feel free to free() it, if you don't need it any more.
 *
 * If there's a non printable character in your string, this function will print out an
 * error message at stderr and skip that character. Use papertape_font_string_is_printable
 * if you want to check wheter your string contains non printable characters.
 *
 * Internally speaken, this is the only read access function to the internal single
 * linked font character list.
 *
 **/
byte_t* papertape_font_get_label(PAPERTAPE_FONT* font, const char* string, int* size) {
	PAPERTAPE_FONT_CHAR* character_spacing = papertape_font_low_get_special(font, PAPERTAPE_FONT_CHARACTER_SPACING_NAME);
	int x;       // to count throught the string
	*size = 0;   // the size of the output byte array
	byte_t* output;
	int output_pos;  // current position in output array
	int clen = strlen(string)*2 - 1; // length of characters stack array
	// dynamic stack arrays are ISO C99, compare http://gcc.gnu.org/onlinedocs/gcc-4.3.2/gcc/Variable-Length.html
	PAPERTAPE_FONT_CHAR *characters[ clen ];
	int cpos = 0; // current position in character array
	
	// pack the font char array together
	for(x=0; string[x] != '\0'; x++) {
		// find white spaces
		if(string[x] == ' ')
			characters[cpos] = papertape_font_low_get_special(font, PAPERTAPE_FONT_WHITE_SPACE_NAME);
		else
			characters[cpos] = papertape_font_low_get_char(font, string[x]);

		if(characters[cpos] == NULL) {
			fprintf(stderr, "papertape_font_get_label: Character \"%c\" at position %i in \"%s\" not printable\n",
				string[x], x+1, string);
			// tell the malloc loop (see below) to loop not over the complete
			// characters[] array, because one character is missing. That is,
			// pop the array for 1 value.
			clen--;
			// don't increase cpos, we want to overwrite this character in the
			// next loop.
			continue;
		} else if(characters[cpos]->bytes == NULL) {
			// for all possibilities... this character is empty.
			// Make sure bytes_n is it, too.
			if(characters[cpos]->bytes_n != 0)
				fprintf(stderr, "papertape_font_get_label: Bad font parsed! Character 0x%x (%c) not really empty!\n",
					characters[cpos]->ascii_id, characters[cpos]->ascii_id);
		}

		*size += characters[cpos]->bytes_n;
		cpos++;
		
		// add the characters spacing after each character
		if(x+1 < strlen(string)) { // but not at the last character
			characters[cpos] = character_spacing;
			*size += characters[cpos]->bytes_n;
			cpos++;
		}
	}
	
	// now malloc the output byte array and generate it
	output = (byte_t*) malloc(*size);
	for(cpos=0,output_pos=0; cpos < clen && output_pos < *size; output_pos += characters[cpos++]->bytes_n)
		memcpy(output+output_pos, characters[cpos]->bytes, characters[cpos]->bytes_n);

	// we are done.
	return output;
}

/**
 * Checks if there's any non printable character in your string. Papertape fonts don't have
 * to have bit masks for every character, so use this function to check your string.
 * papertape_font_get_label will output errors to stderr if it finds unprintable
 * characters.
 *
 * @returns 1 if all characters are printable, 0 otherwise.
 **/
int papertape_font_string_is_printable(PAPERTAPE_FONT *font, const char* string) {
	int x;

	for(x=0; string[x] != '\0'; x++) {
		if(string[x] == ' ')
			// white spaces are always printable
			continue;
		if(papertape_font_low_get_char(font, string[x]) == NULL)
			// character not found!
			return 0;
	}

	return 1; // No non printable char found.
}

/**
 * Write the current PAPERTAPE_FONT (back) out to a file.
 * FILE mus be a writeable handle.
 **/
int papertape_font_write_to_file(PAPERTAPE_FONT* font, FILE *fh) {
	PAPERTAPE_FONT_CHAR* current;
	int x;

	// Write first lines.
	fputs(PAPERTAPE_FONT_WRITE_FIRST_LINES, fh);
	
	// now go throught list and write all characters out.
	current = font;
	while(1) {
		// write bytes and identifiers
		for(x=0; x<current->bytes_n; x++) {
			papertape_font_byte2line(current->bytes[x], fh);
			if(current->ascii_id == 0) {
				fputs(PAPERTAPE_FONT_SPECIAL_ID, fh);
				fputs(current->name, fh);
			} else
				fputc(current->ascii_id, fh);
			fputc('\n', fh);
		}
	
		// and go on. Or break, if at end.
		if(current->next != NULL) {
			// write a neat NULL byte as eye candy seperation
			papertape_font_byte2line(0x0, fh);
			fputc('\n', fh);
			current = current->next;
		} else
			break;
	}
	
	return 1; // could be void.
}
/**
 * This will print a dump of the papertape font linked list, just for debugging purpose,
 * to the file handle. It's just fine to give "stderr" at the file handle.
 **/
void papertape_font_dump(PAPERTAPE_FONT *font, FILE* fh) {
	PAPERTAPE_FONT_CHAR* current;
	int ram_bytes = 0;
	int i = 0;
	int x;

	current = font;
	while(1) {
		ram_bytes += sizeof(PAPERTAPE_FONT_CHAR);
		if(current->ascii_id == 0 && current->name == NULL)
			fprintf(fh, "%i. char: Special character, no name (broken), %i bytes allocated\n", i++, current->bytes_n);
		else if(current->ascii_id == 0) {
			fprintf(fh, "%i. char: Special character, name \"%s\", %i bytes allocated\n", i++, current->name, current->bytes_n);
			ram_bytes += strlen(current->name)+1;
		} else
			fprintf(fh, "%i. char: \"%c\" (0x%x), %i bytes allocated\n", i++, current->ascii_id, current->ascii_id, current->bytes_n);
		if(current->bytes == NULL) {
			fprintf(fh, "No bytes allocated!\n");
		} else {
			ram_bytes += current->bytes_n;
			for(x=0; x < current->bytes_n; x++) {
				papertape_font_byte2line(current->bytes[x], fh);
				fputc('\n', fh);
			}
		}
		
		if(current->next == NULL) {
			fprintf(fh, "End of list.\n");
			fprintf(fh, "Statistics: %i entries, %i total bytes in heap\n", i, ram_bytes);
			return;
		} else
			current = current->next;
	}
}

/** 
 * The main() method (and auxillary helper functions) is only compiled when the switch
 * "STANDALONE" is defined. If not, we'll only get the simple library.
 *
 **/
#ifdef STANDALONE

/**
 * A helper method for the standalone main(): Read contents of a stream
 * in a dynamically allocated heap byte array.
 **/
int file_get_contents(FILE *stream, byte_t **content) {
	byte_t buf[4096];
	size_t bytes; // gerade eben eingelesene bytes
	byte_t *str = NULL;
	size_t total_bytes = 0;     // alle bis jetzt eingelesenen bytes
	size_t total_allocated = 0;
	byte_t *tmp; // fuers realloc

	while(!feof(stream)) {
		bytes = fread(buf, 1, sizeof(buf), stream);

		while( (total_bytes + bytes) > total_allocated) {
			if(str)
				total_allocated *= 2;
			else
				total_allocated = bytes > sizeof(buf) ? bytes : sizeof(buf);

			tmp = realloc(str, total_allocated);

			if(tmp == NULL) {
				fprintf(stderr, "*** file_get_contents ERROR*** Could not reallocate\n");
				//return length; // Fehler - das eingelesene zumindest zurueckgeben.
				// nee, gar nichts zurckgeben. Das geht so nicht.
				return 0;
			}

			str = tmp;
		} // while innen

		memcpy(str + total_bytes, buf, bytes);
		total_bytes += bytes;
	} // while

	if(total_allocated == 0)
		str = (byte_t*) malloc(1); // something empty. Just to be not NULL...
	//str[total_bytes] = '\0'; // wenns ein string wre.

	*content = str;
	return total_bytes;
}

/** 
 * The main method. We are using an own parameter system with extensive help messages.
 * No so much parameter flexibility, but it should be enough :-)
 **/
int main(int argc, char **argv) {
	// Instead of poor getopt emulation, an own parameter system like "svn" CLI client has it
	if(argc < 3 || strcasecmp(argv[1], "help") == 0) {
		// argc == 1: Wrong call like "./font", missing subcommand.
		// Wrong call (like "./font label"), or help call (like "./font help").
		// Display help. Or startup help system.
		if(argc < 3) {
			fprintf(stderr,	"Usage: %s <subcommand> [subcommand option] <font file>\n"
				"The Papertape font utility program.\n"
				"Type '%s help <subcommand>' for help on a specific subcommand.\n"
				"The <subcommand> and <font file> parameters are mandatory\n"
				"\n"
				"Available subcommands:\n"
				"   help        Starts up this help system\n"
				"   label       Prints out a complete label, given as the option\n"
				"   set         Sets the character with the ASCII ID (e.g. 0x20), reads from STDIN\n"
				"   del         Deletes the character with the ASCII ID from the font file\n"
				"   dump        Parses the font file and prints out the result, for debugging\n"
				"   export      Exports the font file to STDOUT, according to schematics\n"
				"   import      Imports the font file, according to schematics\n"
				"               Default schematics: %s\n",
				argv[0], argv[0], PAPERTAPE_FONT_SCHEMATICS_DEFAULT);
		}
				
		if(argc >= 3) {
			// verbose Help system.
			if(strcasecmp(argv[2], "label") == 0) {
				fprintf(stderr,	"label: Print out the compiled bytes for the translated label.\n"
						"Expects String as subcommand option.\n\n"
						"Example calls:\n"
						" Prints out yor text to STDOUT.\n"
						"   $ %s label \"The text you want to label\" your-font-file.ptf:\n"
						" The same with \"hello world\". Upper/lowercase produces the same result, btw.\n"
						"   $ %s label hello\\ world your-font-file.ptf:\n"
						" This writes the output to the file hallo-welt.bin:\n"
						"   $ %s label hallo\\ Welt your-font-file.ptf > hallo-welt.bin\n"
						" This will punch out all files in the directory with the right label before\n"
						"   $ for x in *; do %s label \"$x\" ../fonts/your.ptf | cat - \"$x\" | ./punch; done;\n",
						argv[0], argv[0], argv[0], argv[0]);
			} else if(strcasecmp(argv[2], "set") == 0) {
				fprintf(stderr,	"set: Changes or adds a character in the font file\n"
						"Expects the wanted character as hex or character as subcommand option.\n\n"
						"Example calls:\n"
						" This will create a new font file that contains the character 'a':\n"
						"   $ cat how-a-should-look-like.bin | %s set a new-font-file.ptf\n"
						" This will overwrite the current value of the character 'a':\n"
						"   $ cat a-new-look-for-letter-a.bin | %s set a new-font-file.ptf\n"
						" This will set the blank space character:\n"
						"   $ echo -en \"\\0\" | %s set \" \" existing-font-file.ptf\n"
						" This will do exactly the same like above:\n"
						"   $ echo -en \"\\0\" | %s set 0x20 existing-font-file.ptf\n",
						argv[0], argv[0], argv[0], argv[0]);
			} else if(strcasecmp(argv[2], "del") == 0) {
				fprintf(stderr,	"del: Deletes one or more characters from a font file\n"
						"Expects one character as subcommand option.\n\n"
						"Example calls:\n"
						" Delete a 'm' from your font file:\n"
						"   $ %s del m your-font-file.ptf\n",
						argv[0]);
			} else if(strcasecmp(argv[2], "dump") == 0) {
				fprintf(stderr, "dump: Parse a font file and print out the results to stdout. This is\n"
						"pracitcal for detecting syntax errors in the font file.\n\n"
						"Example calls:\n"
						"  Get a dump:\n"
						"   $ %s dump your-font-file.ptf\n"
						"  Save it somewhere:\n"
						"   $ %s dump your-font-file.ptf > dump.txt\n",
						argv[0], argv[0]);
			} else if(strcasecmp(argv[2], "export") == 0|| strcasecmp(argv[2], "import") == 0) {
				fprintf(stderr,	"import/export: Converting font files to complete binary files on the fly\n"
						"You can give a scheme optionally as subcommand option.\n"
						"This is the default scheme:\n"
						"   %s\n\n"
						"Example calls:\n"
						" Export an existing font file, edit the file with an hex editor and import it again:\n"
						"   $ %s export your-font-file.ptf > font-file-export.bin\n"
						"   $ khexdump font-file-export.bin\n"
						"          [ some GUI editing, and so on...]\n"
						"   $ cat font-file-export.bin | %s import your-font-file.ptf\n"
						" Import to a new font file from existin g other font mechanism:\n"
						"   $ SCHEME=\"abcdefgh...4567890 ()<>\"\n"
						"   $ perl another-font.pl \"$SCHEME\" | %s import \"%s\" importet-font-file.ptf\n"
						" Make a new font file which only contains the digits of the old file:\n"
						"   $ %s export \"0123456789\" existing.ptf | %s import \"0123456789\" new.ptf\n",
						PAPERTAPE_FONT_SCHEMATICS_DEFAULT, argv[0], argv[0], argv[0], argv[0], argv[0], argv[0]);
			} else {
				fprintf(stderr, "'%s': unknown subcommand. Type '%s help' for a list of valid commands.\n",
					argv[2], argv[0]);
			}
		} // verbose help
		return 0;
	} // Help messages

	if(strcasecmp(argv[1], "export") == 0 || strcasecmp(argv[1], "label") == 0) {
		// export or label.
		int size;
		byte_t *bytes;
		PAPERTAPE_FONT *font;
		 
		if(! (font=papertape_font_new_from_file(argc == 4 ? argv[3] : argv[2]) )) {
			fprintf(stderr, "Error creating Papertape font.\n");
			return 3;
		}

		bytes = papertape_font_get_label(font, argc == 4 ? argv[2] : PAPERTAPE_FONT_SCHEMATICS_DEFAULT, &size);
		fwrite(bytes, size, sizeof(byte_t), stdout);	
		return 0;
	} else if(strcasecmp(argv[1], "set") == 0 || strcasecmp(argv[1], "del") == 0) {
		// set and delete
		char character;
		FILE *font_file;
		PAPERTAPE_FONT *font;
		
		// at first: get the character
		if(argc == 3) {
			fprintf(stderr, "%s: Too few arguments! Missing character. Type '%s help %s' for usage.\n",
				argv[1], argv[0], argv[1]);
			return 1;
		}
		if(strlen(argv[2]) > 1)
			// A Number like "255", 0x20 or 080. strol will parse it.
			character = (char)strtol(argv[2], NULL, 0);
		else
			// only one character in ASCII.
			character = argv[2][0];

		// open file for reading and WRITING.
		font_file = fopen(argv[3], "r+");
		if(!font_file) {
			perror("Opening font file for writing");
			return 2;
		}
		font = papertape_font_new();
		papertape_font_load_file(font, font_file);

		if(strcasecmp(argv[1], "set")) {
			// change/add character
			byte_t *contents;
			int bytes = file_get_contents(stdin, &contents);
			printf("Read in %i bytes for letter %c (seen as byte: 0x%x)\n", bytes, character, character);
			papertape_font_set_char(font, character, contents, bytes);
			printf("Set character %c successfully.\n", character);
		} else {
			// delete character
			printf(papertape_font_del_char(font, character) ? 
				"Deleted %c successfully from file %s\n" :
				"Could not find %c in file %s!\n",
				character, argv[3]);
		}
		
		// write papertape file.
		papertape_font_write_to_file(font, font_file);
		return 0;
	} else if(strcasecmp(argv[1], "dump") == 0) {
		PAPERTAPE_FONT* font = papertape_font_new_from_file(argv[2]);
		printf("This is a dump from the internal single linked ordered list,\n"
			"created from the file '%s':\n\n", argv[2]);
		papertape_font_dump(font, stdout);
		return 0;
	} else if(strcasecmp(argv[1], "import") == 0) {
		// run an import.
		PAPERTAPE_FONT *font = papertape_font_new();
		FILE* font_file;
		byte_t *contents;
		int content_len;
		char *schematics;
		
		// open file for DESTROYING WRITING.
		font_file = fopen(argc == 4 ? argv[3] : argv[2], "w");
		if(font_file == NULL) {
			perror("Opening font file for writing");
			return 2;
		}
		
		schematics = argc == 4 ? argv[2] : PAPERTAPE_FONT_SCHEMATICS_DEFAULT;

		fprintf(stderr, "Reading from stdin, expecting schematics %s\n", schematics);
		content_len = file_get_contents(stdin, &contents);
		// debug: print it out!
		// fwrite(contents, 1, content_len, stdout); exit(0);
	
		// Parse schematics
		papertape_font_load_schematics(font, schematics, contents, content_len);

		fprintf(stderr, "Readed from stdin successfully. Here you get a dump:\n");
		papertape_font_dump(font, stdout);

		// that's it, we are done. Write the papertape font.
		papertape_font_write_to_file(font, font_file);
		return 0;
	} else {
		fprintf(stderr, "'%s': unknown subcommand. Type '%s help' for a list of valid commands.\n",
			argv[2], argv[0]);
		return 1;
	}
} // main
#endif /* STANDALONE */
